/* Tests the throughput of the link to a tserver program. See the
  ThroughputServer.cpp file for details.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//_______________ Static Variables _______________

static DEngine netEngine;//handles network events
static DClient *client = NULL;//handles our connection to the server
const uint16_t PORT_NUMBER = 5127;//The port number to communicate over
static double kbReceived = 0;//The total number of kilobytes received
static int receiveBufferSize = 0;//The number of kilobytes to allocate in our receive buffer
static bool allOK = true;//False if we encountered an error and should stop the program

//Shows tclient's command-line usage
void showUsage(void);

//_______________ Event Handlers _______________

//Called when the connect operation finishes
void connectHandler(string message,
                    libdasyne::NetworkCode status)
{
    if (status == libdasyne::DA_SUCCESS)
    {
        cout << "Connected to " << client->getRemoteAddress() << endl;
        cout << "My address: " << client->getLocalAddress() << endl;
    }//if status
    else
    {
        cout << message << endl;
        allOK = false;
    }//else status
}//connectHandler

//Event handler, called when we receive data
void receiveHandler(DBuffer &receivedBytes)
{
    kbReceived += receivedBytes.size() / 1024.0;
}//receiveHandler

//Event handler, called when a network error occurs
void errorHandler(string message,
                  bool socketOK)
{
    cout << message << endl;

    if (!socketOK)
    {
        cout << "Connection lost." << endl;
        allOK = false;
    }//if !socketOK
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Verify that the proper number of arguments were given ###
    if (argc != 3)
    {
        showUsage();
        allOK = false;
    }//if argc

    //### Extract the buffer size parameter ###
    if (allOK)
    {
        if (!DEngine::strToInt(argv[2], receiveBufferSize))
        {
            showUsage();
            allOK = false;
        }//if !strToInt()
    }//if allOK

    //### Create the client and connect ###
    if (allOK)
    {
        if (receiveBufferSize > 0)
        {
            client = new DClient(&netEngine, receiveBufferSize * 1024);
            client->setErrorHandler(&errorHandler);
            client->setReceiveHandler(&receiveHandler);
            client->startConnect(argv[1], PORT_NUMBER, DA_IPV4, &connectHandler);
        }//if receiveBufferSize
        else
        {
            showUsage();
            allOK = false;
        }//else receiveBufferSize
    }//if allOK

    while (allOK)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine.update(-1);
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            allOK = false;
        }//if eventsRun

        //### Sleep if we have nothing to do ###
        #ifndef DA_WINDOWS_VERSION
            else if (eventsRun == 0)
                DEngine::sleep(1);
        #endif
    }//while allOK

    //### Output results ###
    if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1048576.0 << " GB" << endl;
    else if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1024.0 << " MB" << endl;
    else
        cout << "Received " << kbReceived << " KB" << endl;

    //### Delete the classes that use netEngine before netEngine is deleted ###
    delete client;

    return 0;
}//main

void showUsage(void)
{
    cout << "Usage: tclient <host_name> <buffer_KB>" << endl;
    cout << endl;
}//showUsage
