/*
 * Copyright (C) 2005-2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
  A deque of uint8_t arrays, meant to be used with scatter-gather IO. TcpClient provides a
  send function where bytes are added to the end of a queue when the function is called, and
  removes bytes from the front when the bytes are sent over the network. This allows the programmer
  to send as much as he likes without needing to worry about buffer sizes, or re-sending data that
  wasn't sent.

  However, a deque of bytes is slow because we can only read or pop_front() one byte at a time. DBytes
  providesa deque of buffers, where each element in the deque is an array. This allows rapid copies to
  the TcpClient, and we can use scatter/gather IO so that the socket's send function doesn't require
  bytes to be copied at all. (ASIO sockets don't accept deques, but they will accept a vector of arrays)

  DBytes is rather complex and dangerous, because it needs to do elaborate footwork for simple
  operations, and uses unbounded arrays. If there are any bugs in Dasyne, they're likely to be
  here!

  Performance repeatedly sending 64K of data to 127.0.0.1 on my dual-core workstation:

    Deque:                    32.8867 MB/second
    Deque, no bounds-check:   44.2347 MB/second
    DBytes, array size 4K:    216.566 MB/second

  We get a speedup of 4.9 times compared to a deque without bounds checking.
*/

#ifndef DBYTES_HPP_INCLUDED
#define DBYTES_HPP_INCLUDED

#include <iostream>
#include <fstream>
#include <string>
#include <deque>

#include "DEngine.hpp"

namespace libdasyne
{
    class DSafeArray;

    class DBytes
    {
        public:
            //constructor
            DBytes(void);

            //copy constructor
            DBytes(const DBytes &source);

            //destructor
            virtual ~DBytes(void);

            //Overloaded = operator
            DBytes& operator= (const DBytes &source);

            //returns the character at the specified location
            uint8_t getByte(long int pos) const;

            //changes the byte at the specified index, does nothing if the index is past
            //the end of the list
            void setByte(long int pos,
                         uint8_t newByte);

            //returns the number of characters in byteData
            unsigned long int size(void) const;

            //delete all entries in byteData
            void clear(void);

            //Remove the specified number of bytes from the front of byteData. Clears byteData
            //if it doesn't contain numRemovals bytes
            void removeFirst(long int numRemovals);

            //insert source at the front or back of the DBytes
            void appendByte(uint8_t source);
            void insertByte(uint8_t source);

            //append source to the end of this DBytes
            void append(const DBytes &source);
            void append(const DBytes &source,
                        long int index,
                        long int length);
            void appendArray(const uint8_t *source,
                             long int length);

            //Copies the contents of the DBytes into theArray, copying up to arraySize bytes
            void toArray(uint8_t *theArray,
                         uint64_t arraySize);

            //Returns the contents of this DBytes as a series of ASIO buffers, intended to be used
            //with a socket send operation
            std::vector<boost::asio::const_buffer> getSendBuffers(long int maxBytes) const;

            //Returns the number of DBytes in existence, used to check for memory leaks.
            //Always returns 0 if DA_DEBUG_MODE is not defined
            static long int getNumInstances(void);

            //The number of bytes in each element of byteData
            static const int ARRAY_SIZE = 4096;

            //Allow the DBufferIter class to access our private functions
            friend class DBufferIter;

        private:
            //The data we're storing. Each element is a pointer to an array of ARRAY_SIZE bytes,
            //for use with scatter-gather IO
            #ifdef DA_DEBUG_MODE
                std::deque<DSafeArray> byteData;

                //The current number of DBytes instances
                static long int numInstances;

                //The number of SafeArrays added to byteData
                long int arraysAdded;
            #else
                std::deque <uint8_t*> byteData;
            #endif

            //The number of bytes being used at the first and last elements of theData
            //If there is only one element, startFilledBytes is used and endFilledBytes is zero.
            uint16_t startFilledBytes;
            uint16_t endFilledBytes;

            //############### FUNCTIONS ###############

            //given a byte position, calculate the equivalent byteData and array indexes
            void calcIndexes(long int pos,
                             uint32_t &dataIndex,
                             unsigned int &arrayIndex) const;

            //Adds a new element to the end of byteData, sets endFilledBytes to zero
            void addElement(void);

            //Called by removeFirst when a deque with multiple elements is reduced to one element.
            //We need to switch from endFilledBytes to startFilledBytes, and shuffle the array items
            //to start at the end of the array rather than the top.
            void convertLastToFirst(long int numRemovals);

            //Sets theArrays to the data being held by byteData. Only adds maxBytes bytes
            //worth of data to theArrays. This is a shallow copy.
            void getArrays(std::vector<uint8_t*> &theArrays,
                           std::vector<unsigned int> &arraySizes,
                           uint64_t maxBytes) const;

    };//DBytes
}//namespace libdasyne

#endif
