#include "DDispatcher.hpp"
#include "DEventMap.hpp"
#include "DBuffer.hpp"
#include "DBufferIter.hpp"
#include "DFlushEvent.hpp"

using namespace std;
using namespace libdasyne;

DDispatcher::DDispatcher(DEngine *newNetEngine) : DClient(newNetEngine)
{
    initVars();
}//constructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         unsigned int newBufferSize) : DClient(newNetEngine, newBufferSize)
{
    initVars();
}//cosntructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         boost::asio::ip::tcp::socket *newSocket) : DClient(newNetEngine, newSocket)
{
    initVars();
}//constructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         boost::asio::ip::tcp::socket *newSocket,
                         unsigned int newBufferSize) : DClient(newNetEngine, newSocket, newBufferSize)
{
    initVars();
}//constructor

DDispatcher::~DDispatcher(void)
{
    for (unsigned int i = 0; i < subtypeMaps.size(); i++)
        delete subtypeMaps.at(i);
    subtypeMaps.clear();

    clearOutMessages();
}//destructor

void DDispatcher::initVars(void)
{
    myState = DADD_FIRST_HEADER;
    recvSize = 0;
    eventHandler = NULL;
    streamOK = true;
    junkAmount = 0;
    haveTypes = false;
    haveSingle = false;
    sizeFieldLength = 0;

    for (unsigned int i = 0; i < 256; i++)
        subtypeMaps.push_back(new DEventMap(i));

    setReceiveHandler(boost::bind(&DDispatcher::receiveData, this, _1));
}//initVars

void DDispatcher::dispatch(DBuffer *newMessage)
{
    //### Send the message immediately ###
    if (getDelayEnabled())
    {
        DClient::send(*newMessage);

        delete newMessage;
        newMessage = NULL;
    }//if getDelayEnabled()

    //### Wait for a flush before sending ###
    else
    {
        outMessages.push_back(newMessage);
        newMessage = NULL;
    }//else getDelayEnabled()
}//dispatch

void DDispatcher::sendMessage(const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    appendHeader(theData, newMessage);

    newMessage->append(theData);
    dispatch(newMessage);
}//sendMessage

void DDispatcher::sendMessage(uint8_t messageType,
                              const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    appendHeader(theData, newMessage);

    newMessage->appendByte(messageType);
    newMessage->append(theData);

    dispatch(newMessage);
}//sendMessage

void DDispatcher::sendMessage(uint8_t messageType,
                              uint16_t subType,
                              const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    appendHeader(theData, newMessage);

    newMessage->appendByte(messageType);
    newMessage->appendVariable(subType);
    newMessage->append(theData);

    dispatch(newMessage);
}//sendMessage

void DDispatcher::appendHeader(const DBuffer &theData,
                               DBuffer *newMessage)
{
    uint32_t fourByteSize = 0;//the maximum message size that fits in four bytes

    //### Calculate the max message size that fits in four bytes ###
    fourByteSize = 2147483647;
    fourByteSize *= 2;
    fourByteSize++;

    //### Add three bytes of header data ###
    newMessage->appendByte(FIRST_HEADER_BYTE);
    newMessage->appendByte(SECOND_HEADER_BYTE);
    newMessage->appendByte(THIRD_HEADER_BYTE);

    //### Add the message size ###
    if (theData.size() == 0)
        newMessage->appendByte(0);

    else if (theData.size() <= 255)
    {
        newMessage->appendByte(1);
        newMessage->appendVariable(static_cast<uint8_t>(theData.size()));
    }//else if size()

    else if (theData.size() <= 65535)
    {
        newMessage->appendByte(2);
        newMessage->appendVariable(static_cast<uint16_t>(theData.size()));
    }//else if size()

    else if (theData.size() <= fourByteSize)
    {
        newMessage->appendByte(4);
        newMessage->appendVariable(static_cast<uint32_t>(theData.size()));
    }//else if size()

    else
    {
        newMessage->appendByte(8);
        newMessage->appendVariable(static_cast<uint64_t>(theData.size()));
    }//else size()
}//appendHeader

void DDispatcher::addHandler(boost::function<void (DBuffer&)> newHandler)
{
    if (haveTypes)
    {
        DEngine::addFatal("Tried to add a generic event handler, but we already ");
        DEngine::fatalError("have handlers for specific message types.");
    }//if haveTypes

    singleHandler = newHandler;
    haveSingle = true;
}//addHandler

void DDispatcher::addHandler(uint8_t messageType,
                             boost::function<void (DBuffer&)> newHandler)
{
    if (haveSingle)
    {
        DEngine::addFatal("Tried to add a handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but we already have a handler for all types of messages.");
    }//if haveSingle

    subtypeMaps.at(messageType)->addHandler(newHandler);
    haveTypes = true;
}//addHandler

void DDispatcher::addHandler(uint8_t messageType,
                             uint16_t subType,
                             boost::function<void (DBuffer&)> newHandler)
{
    if (haveSingle)
    {
        DEngine::addFatal("Tried to add a handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but we already have a handler for all types of messages.");
    }//if haveSingle

    subtypeMaps.at(messageType)->addHandler(subType, newHandler);
    haveTypes = true;
}//addHandler

void DDispatcher::receiveData(libdasyne::DBuffer &receivedBytes)
{
    DBufferIter recvIter;//iterates through the received bytes
    unsigned int byteIndex = 0;//the index of the current byte to extract from receivedBytes
    int copyAmount = 0;//the number of bytes to copy from receivedBytes
    int bytesAvailable = 0;//the number of bytes remaining in receivedBytes
    uint16_t temp = 0;//stores a two-byte parameter from receivedBytes
    uint8_t oneByte = 0;//a single byte from receivedBytes

    if (!streamOK)
        junkAmount += receivedBytes.size();

    recvIter.setBuffer(&receivedBytes);

    while (!recvIter.finished())
    {
        oneByte = recvIter.value();

        switch (myState)
        {
            case DADD_FIRST_HEADER:
                if (checkHeaderByte(FIRST_HEADER_BYTE, oneByte))
                    myState = DADD_SECOND_HEADER;
            break;

            case DADD_SECOND_HEADER:
                if (checkHeaderByte(SECOND_HEADER_BYTE, oneByte))
                    myState = DADD_THIRD_HEADER;
            break;

            case DADD_THIRD_HEADER:
                if (checkHeaderByte(THIRD_HEADER_BYTE, oneByte))
                {
                    myState = DADD_SIZE_LENGTH;
                    streamOK = true;
                    junkAmount = 0;
                }//if checkHeaderByte
            break;

            case DADD_SIZE_LENGTH:
                sizeFieldLength = oneByte;

                if (sizeFieldLength == 0)
                {
                    recvSize = 0;
                    recvMessage.clear();

                    if (haveSingle)
                    {
                        singleHandler(recvMessage);
                        myState = DADD_FIRST_HEADER;
                    }//if haveSingle
                    else
                       myState = DADD_MESSAGE_TYPE;
                }//if sizeFieldLength

                else if ((sizeFieldLength != 1)
                      && (sizeFieldLength != 2)
                      && (sizeFieldLength != 4)
                      && (sizeFieldLength != 8))
                {
                    errMsg = "Invalid size for the length value: ";
                    errMsg.append(DEngine::intToStr(sizeFieldLength));
                    reportError(errMsg, true);
                    myState = DADD_FIRST_HEADER;
                }//if sizeFieldLength ... && sizeFieldLength

                else
                {
                    dataField.clear();
                    myState = DADD_SIZE;
                }//else sizeFieldLength ... && sizeFieldLength
            break;

            case DADD_SIZE:
                extractMessageLength(oneByte);
            break;

            case DADD_MESSAGE_TYPE:
                if (subtypeMaps.at(oneByte)->getHaveSingle())
                {
                    currentHandler = subtypeMaps.at(oneByte)->getSingleHandler();

                    if (recvSize == 0)
                    {
                        currentHandler(recvMessage);
                        myState = DADD_FIRST_HEADER;
                    }//if recvSize
                    else
                        myState = DADD_MESSAGE_BODY;
                }//if getHaveSingle()

                else if (subtypeMaps.at(oneByte)->size() > 0)
                {
                    recvMsgType = oneByte;
                    dataField.clear();
                    myState = DADD_FIRST_SUBTYPE;
                }//else if size()

                else
                {
                    errMsg = "No handler set for message type ";
                    errMsg.append(DEngine::intToStr(oneByte));
                    reportError(errMsg, true);
                    myState = DADD_FIRST_HEADER;
                    streamOK = false;
                }//else size()
            break;

            case DADD_FIRST_SUBTYPE:
                dataField.appendByte(oneByte);
                myState = DADD_SECOND_SUBTYPE;
            break;

            case DADD_SECOND_SUBTYPE:
                dataField.appendByte(oneByte);
                dataField.extractVariable(0, temp);

                if (subtypeMaps.at(recvMsgType)->getHandler(temp, currentHandler))
                {
                    if (recvSize == 0)
                    {
                        currentHandler(recvMessage);
                        myState = DADD_FIRST_HEADER;
                    }//if recvSize
                    else
                        myState = DADD_MESSAGE_BODY;
                }//if getHandler()
                else
                {
                    errMsg = "No handler set for message type ";
                    errMsg.append(DEngine::intToStr(recvMsgType));
                    errMsg.append(" sub-type ");
                    errMsg.append(DEngine::intToStr(temp));
                    reportError(errMsg, true);

                    myState = DADD_FIRST_HEADER;
                    streamOK = false;
                }//else getHandler
            break;

            case DADD_MESSAGE_BODY:
                //### Calculate the number of bytes to copy ###
                copyAmount = recvSize - recvMessage.size();
                if (copyAmount <= 0)
                    DEngine::fatalError("DDispatcher::receiveData(): received too many bytes!");

                //### Reduce it to the number of bytes available ###
                bytesAvailable = receivedBytes.size() - byteIndex;
                if (bytesAvailable < copyAmount)
                    copyAmount = bytesAvailable;

                //### Copy the bytes ###
                recvMessage.append(receivedBytes, byteIndex, copyAmount);
                byteIndex += copyAmount - 1;
                recvIter.setIndex(byteIndex);

                //### Report that we have a complete message ###
                if (recvMessage.size() == recvSize)
                {
                    myState = DADD_FIRST_HEADER;
                    currentHandler(recvMessage);
                }//if size()
                else if (recvMessage.size() > recvSize)
                    DEngine::fatalError("DDispatcher::receiveData(): stored too many bytes?!");
            break;

            default:
                DEngine::fatalError("DDispatcher::receiveData(): reached an unknown state");
                break;
        };//switch myState

        recvIter.next();
        byteIndex++;
    }//while !finished()

    if (!streamOK)
    {
        while (junkAmount > 65541)//64436 + 5 bytes header size
        {
            junkAmount -= 65541;
            errMsg = "Received 64 KB of data without finding a valid message.";
            reportError(errMsg, true);
        }//while junkAmount
    }//if !streamOK
}//receiveData

void DDispatcher::extractMessageLength(uint8_t oneByte)
{
    uint8_t size8 = 0;//store the message length in a variable of the appropriate size
    uint16_t size16 = 0;
    uint32_t size32 = 0;
    uint64_t size64 = 0;

    dataField.appendByte(oneByte);
    if (dataField.size() >= sizeFieldLength)
    {
        if (sizeFieldLength == 1)
        {
            dataField.extractVariable(0, size8);
            recvSize = size8;
        }//if sizeFieldLength

        else if (sizeFieldLength == 2)
        {
            dataField.extractVariable(0, size16);
            recvSize = size16;
        }//else if sizeFieldLength

        else if (sizeFieldLength == 4)
        {
            dataField.extractVariable(0, size32);
            recvSize = size32;
        }//else if sizeFieldLength

        else if (sizeFieldLength == 8)
        {
            dataField.extractVariable(0, size64);
            recvSize = size64;
        }//else if sizeFieldLength

        else
            DEngine::fatalError("DDispatcher::extractMessageLength(): invalid sizeFieldLength");

        recvMessage.clear();
        currentHandler = NULL;

        if (haveSingle)
        {
            currentHandler = singleHandler;
            myState = DADD_MESSAGE_BODY;
        }//if haveSingle
        else
            myState = DADD_MESSAGE_TYPE;
    }//if size()
}//extractMessageLength

bool DDispatcher::checkHeaderByte(uint8_t expected,
                                  uint8_t received)
{
    bool result = true;//false if we received an invalid header byte

    if (received != expected)
    {
        myState = DADD_FIRST_HEADER;
        result = false;

        if (streamOK)
        {
            errMsg = "Expected header byte ";
            errMsg.append(DEngine::intToStr(expected));
            errMsg.append(", got ");
            errMsg.append(DEngine::intToStr(received));

            reportError(errMsg, true);
            streamOK = false;
        }//if streamOK
    }//if received

    return result;
}//checkHeaderByte

void DDispatcher::delayToggled(void)
{
    if (getDelayEnabled())
    {
        delete eventHandler;
        eventHandler = NULL;
    }//if getDelayEnabled
    else
    {
        if (eventHandler == NULL)
            eventHandler = new DFlushEvent(internalNetEngine, this);
    }//else getDelayEnabled
}//delayToggled

bool DDispatcher::flush(void)
{
    bool result = false;//true if there was data to flush

    if (outMessages.size() > 0)
        result = true;

    for (unsigned int i = 0; i < outMessages.size(); i++)
        DClient::send(*(outMessages.at(i)));
    clearOutMessages();

    return result;
}//flush

unsigned long int DDispatcher::unsentBytes(void)
{
    unsigned long int result = 0;//the number of bytes remaining to send

    result = DClient::unsentBytes();

    for (unsigned long int i = 0; i < outMessages.size(); i++)
        result += outMessages.at(i)->size();

    return result;
}//unsentBytes

void DDispatcher::clearOutMessages(void)
{
    for (unsigned long int i = 0; i < outMessages.size(); i++)
        delete outMessages.at(i);
    outMessages.clear();
}//clearOutMessages
