/*
 * Copyright (C) 2011 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
  Provides message-based communication. Normally, we must deal with the situation where
  only partial messages are obtained from the TCP stream by waiting until more data arrives.
  DMessageClient seperates the stream into distinct messages, and calls messageHandler
  whenever a complete message arrives. This comes at an overhead of 5 bytes per message,
  2 bytes for the message size, and a 3-byte header that allows us to recover if the TCP stream
  is corrupted.

  Furthermore, the DDispatcher allows us to split a single TCP connection into seperate 'channels'.
  We can add a 1-byte type with a 2-byte subtype, giving us over 16 million possible channels. If no
  type is specified, the DDispatcher won't send any type information over the network, and we keep our
  5 bytes of overhead per message. Likewise, if the subtype isn't specified, we have an overhead of 6 bytes,
  5 from the header and one from the type.
*/

#ifndef DDISPATCHER_HPP_INCLUDED
#define DDISPATCHER_HPP_INCLUDED

#include "DEngine.hpp"
#include "DClient.hpp"
#include "DBuffer.hpp"

namespace libdasyne
{
    class DEventMap;
    class DFlushEvent;

    //The states in the finite state machine for parsing received messages
    enum DADD_FSMState
    {
        DADD_FIRST_HEADER, DADD_SECOND_HEADER, DADD_THIRD_HEADER, DADD_SIZE_LENGTH,
        DADD_SIZE, DADD_MESSAGE_TYPE, DADD_FIRST_SUBTYPE, DADD_SECOND_SUBTYPE,
        DADD_MESSAGE_BODY
    };//DAMC_FSMState

    class DDispatcher : private DClient
    {
        public:
            //constructors
            DDispatcher(DEngine *newNetEngine);
            DDispatcher(DEngine *newNetEngine,
                        unsigned int newBufferSize);

            //constructors, should be called when a DServer accepts a new connection
            DDispatcher(DEngine *newNetEngine,
                        boost::asio::ip::tcp::socket *newSocket);
            DDispatcher(DEngine *newNetEngine,
                        boost::asio::ip::tcp::socket *newSocket,
                        unsigned int newBufferSize);

            //destructor
            virtual ~DDispatcher(void);

            //Functions to expose from DClient
            using DClient::startConnect;
            using DClient::closeSocket;
            using DClient::getErrMsg;
            using DClient::getLocalAddress;
            using DClient::getRemoteAddress;
            using DClient::getBufferSize;
            using DClient::setDelay;
            using DClient::setSendHandler;
            using DClient::setErrorHandler;
            using DClient::unsentBytes;
            using DClient::isConnected;

            //Add an event handler to be called when we receive a message of the
            //specified type
            void addHandler(boost::function<void (DBuffer&)> newHandler);
            void addHandler(uint8_t messageType,
                            boost::function<void (DBuffer&)> newHandler);
            void addHandler(uint8_t messageType,
                            uint16_t subType,
                            boost::function<void (DBuffer&)> newHandler);

            //Sends the specified data to the remote host. If Nagle's algorithm has been disabled,
            //the data is not immediately sent; use flush() to send it. This allows you to
            //add many small messages into a single network packet. If flush is not called, the
            //data will be sent the next time DEngine::update() is called.
            void sendMessage(const libdasyne::DBuffer &theData);
            void sendMessage(uint8_t messageType,
                             const libdasyne::DBuffer &theData);
            void sendMessage(uint8_t messageType,
                             uint16_t subType,
                             const libdasyne::DBuffer &theData);

            //Send any buffered data from the send() function. If Nagle's algorithm is enabled,
            //calling flush has no effect, as we are still limited by Nagle's. Returns true if there
            //was any data to be flushed, false otherwise.
            bool flush(void);

            //Returns the number of bytes in the send buffer, plus the number of bytes in outMessages
            unsigned long int unsentBytes(void);

            //Three header bytes that are placed at the beginning of each message.
            //These are used to find the next message if we should lose our place in the
            //TCP stream. The longest message size is 64K. Given random binary data,
            //we have even odds of finding any given 2-byte combination within a message.
            //With three bytes, we have only a 1 in 256 chance of finding a false header,
            //and should be able to get back on track with minimal skipped messages.
            static const int FIRST_HEADER_BYTE = 166;
            static const int SECOND_HEADER_BYTE = 9;
            static const int THIRD_HEADER_BYTE = 57;

        protected:
            //A description of the most recent error
            using DClient::errMsg;

            //Report that setDelay was called, creates or destroys eventHandler as appropriate
            void delayToggled(void);

        private:
            //Each element is an std::map of event handlers. The element's index is the message type,
            //and the map contains the defined subtypes. If there is no subtype, the DEventMap
            //calls a stored event handler.
            std::vector<DEventMap*> subtypeMaps;

            //Each element is a message that should be sent to the server the next time flush is
            //called.
            std::vector<DBuffer*> outMessages;

            //Calls flush() whenever the DEngine updates. Only used if Nagle's algorithm is disabled
            DFlushEvent *eventHandler;

            //Function to call if there is no type for the received message
            boost::function<void (DBuffer&)> singleHandler;

            //Function to call for the message currently being received
            boost::function<void (DBuffer&)> currentHandler;

            //A multi-byte parameter we're extracting from the received message
            DBuffer dataField;

            //The message we're currently receiving
            DBuffer recvMessage;

            //Our current state in the message parsing FSM
            DADD_FSMState myState;

            //The size of the message we're currently receiving
            uint64_t recvSize;

            //The number of bytes we've received since streamOK became false, used to call errorHandler
            //if we've received more than 64K of data without finding a message header.
            long int junkAmount;

            //The type of the message currently being received
            uint8_t recvMsgType;

            //The number of bytes in the message size field
            uint8_t sizeFieldLength;

            //True if we've received a valid message and should call errorHandler if we receive
            //invalid data. Set to false once errorHandler has been called, so that we only call it
            //once per corrupt message.
            bool streamOK;

            //True if we have handlers for different message types
            bool haveTypes;

            //True if singleHandler points to a function, false otherwise
            bool haveSingle;

            //############### FUNCTIONS ###############

            //Initialize member variables, called by the constructors
            void initVars(void);

            //Adds the three-byte header to newMessage
            void appendHeader(DBuffer *newMessage);

            //Appends the 3-byte header and the number of bytes in theData to newMessage.
            //theData should have at most 65536 elements
            void appendHeader(const DBuffer &theData,
                              DBuffer *newMessage);

            //Returns true if expected == received, returns false otherwise and calls errorHandler
            bool checkHeaderByte(uint8_t expected,
                                 uint8_t received);


            //Event handler, called when we've received data
            void receiveData(libdasyne::DBuffer &receivedBytes);

            //Deletes every element in outMessages and clears it
            void clearOutMessages(void);

            //Sends newMessage to the remote host
            void dispatch(DBuffer *newMessage);

            //Extracts the message length from the network stream
            void extractMessageLength(uint8_t oneByte);

    };//DDispatcher
}//namespace libdasyne

#endif
