/*
 * Copyright (C) 2012 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    One or more values of the same type, which form part of a DProtocol message. Slightly clumsier than
    I'd like, due to limitations of C++ templates.
*/

#ifndef DFIELD_HPP_INCLUDED
#define DFIELD_HPP_INCLUDED

#include "DEngine.hpp"
#include "DFieldValue.hpp"

namespace libdasyne
{
    class DField
    {
        public:
            //constructor
            DField(const std::string &newName,
                   FieldType newType);

            //destructor
            ~DField(void);

            //Add a new value to this field. Returns false if we would exceed maxEntries.
            template <class T>
            void addValue(T newValue);
            void addString(const std::string &newValue);
            void addDBuffer(DBuffer &newValue);

            //Retrieve a value from the specified index
            template <class T>
            void getValue(unsigned int index,
                          T &theValue) const;
            void getString(unsigned int index,
                           std::string &theValue) const;
            void getDBuffer(unsigned int index,
                            DBuffer &theValue) const;

            //Returns the number of values being stored in this field.
            uint64_t size(void);

            //Serializes this field, and appends it to the back of theMessage
            void generateMessage(DBuffer &theMessage);

            //Normally, a DField contains exactly one value. setNumValues lets us create a field
            //that can hold multiple values (maxEntries > 1) or a field that is optional
            //(minEntries == 0)
            void setFieldRange(uint64_t newMinEntries,
                               uint64_t newMaxEntries);

            //Remove all values contained in this field
            void clear(void);

            //Extracts this DField from theMessage, increasing byteIndex by the number of bytes
            //consumed. Returns false on errors and sets errorDesc with a description of the problem
            bool parse(const DBuffer &theMessage,
                       unsigned int &byteIndex,
                       std::string &errorDesc);

            //Return's this DField's unique name
            std::string name(void);

        private:
            //The data contained in this field
            std::vector<DFieldValue*> theData;

            //Generates new DFieldValues of the appropriate sub-type
            DFieldValue *generator;

            //A unique name for this DField
            std::string myName;

            //The type of data being stored, eg double, DBuffer*, string, etc.
            FieldType myType;

            //The allowed number of entries in this field
            uint64_t maxEntries;
            uint64_t minEntries;

            //Adds the field name and current byteIndex to errorDesc
            void addFieldDetails(unsigned int byteIndex,
                                 std::string &errorDesc);
    };//DField

    template <class T>
    void DField::addValue(T newValue)
    {
        DFieldValue *newEntry = NULL;//The new item to add to theData
        DBuffer valueBytes;//contains the bytes from newValue

        //### We need to call special handlers for strings and DBuffers ###
        if (myType == DA_STRING)
        {
            DEngine::addFatal("DField::addValue(): ");
            DEngine::addFatal(myName);
            DEngine::fatalError(" contains strings, call addString() instead");
        }//if myType

        else if (myType == DA_DBUFFER)
        {
            DEngine::addFatal("DField::addValue(): ");
            DEngine::addFatal(myName);
            DEngine::fatalError(" contains DBuffers, call addDBuffer() instead");
        }//else if myType

        //### Verify that we aren't exceeding maxEntries ###
        if (theData.size() >= maxEntries)
            DEngine::fatalError("DField::addValue(): would exceed the max number of entries for this field.");

        //### Cast newValue to the appropriate type, lets us create a DBuffer of the right size
        //    even if type T is wrong. (Say T is a float when we want a double)                ###
        switch (myType)
        {
            case DA_UINT8_T:
                valueBytes.appendVariable(static_cast<uint8_t>(newValue));
            break;

            case DA_UINT16_T:
                valueBytes.appendVariable(static_cast<uint16_t>(newValue));
            break;

            case DA_UINT32_T:
                valueBytes.appendVariable(static_cast<uint32_t>(newValue));
            break;

            case DA_UINT64_T:
                valueBytes.appendVariable(static_cast<uint64_t>(newValue));
            break;

            case DA_INT8_T:
                valueBytes.appendVariable(static_cast<int8_t>(newValue));
            break;

            case DA_INT16_T:
                valueBytes.appendVariable(static_cast<int16_t>(newValue));
            break;

            case DA_INT32_T:
                valueBytes.appendVariable(static_cast<int32_t>(newValue));
            break;

            case DA_INT64_T:
                valueBytes.appendVariable(static_cast<int64_t>(newValue));
            break;

            case DA_FLOAT:
                valueBytes.appendVariable(static_cast<float>(newValue));
            break;

            case DA_DOUBLE:
                valueBytes.appendVariable(static_cast<double>(newValue));
            break;

            default:
                DEngine::fatalError("DField::addValue(): unknown field type");
            break;
        }//switch myType

        //### Create the new entry ###
        newEntry = generator->generate();
        newEntry->setValue(valueBytes);
        theData.push_back(newEntry);
    }//addValue

    template <class T>
    void DField::getValue(unsigned int index,
                          T &theValue) const
    {
        DBuffer valueBytes = theData.at(index)->getValue();

        if (myType == DA_STRING)
        {
            DEngine::addFatal("DField::getValue(): ");
            DEngine::addFatal(myName);
            DEngine::fatalError(" contains strings, call getString() instead");
        }//if myType

        else if (myType == DA_DBUFFER)
        {
            DEngine::addFatal("DField::getValue(): ");
            DEngine::addFatal(myName);
            DEngine::fatalError(" contains DBuffers, call getDBuffer() instead");
        }//else if myType

        if (valueBytes.size() != sizeof(T))
        {
            DEngine::addFatal("DField::getValue(): field values have ");
            DEngine::addFatal(DEngine::intToStr(valueBytes.size()));
            DEngine::addFatal(" bytes, but theValue has ");
            DEngine::addFatal(DEngine::intToStr(sizeof(theValue)));
            DEngine::fatalError(" bytes. theValue has a different type than the requested field.");
        }//if size()

        valueBytes.extractVariable(0, theValue);
    }//getValue
}//namespace

#endif
