/*
 * Copyright (C) 2005-2012 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DProtocol.hpp"
#include "DField.hpp"
#include "DBuffer.hpp"

using namespace std;
using namespace libdasyne;

DProtocol::DProtocol(void)
{
}//constructor

DProtocol::~DProtocol(void)
{
    map<string, DField*>::iterator fieldIter;//the current element in fields to delete

    for (fieldIter = fields.begin(); fieldIter != fields.end(); fieldIter++)
        delete fieldIter->second;
    fields.clear();

    protocolOrder.clear();
}//destructor

bool DProtocol::parseMessage(const DBuffer &theMessage)
{
    DField *currentField = NULL;//the current field we're extracting from theMessage
    unsigned int fieldIndex = 0;//the index of the current field to extract
    unsigned int byteIndex = 0;//the index of the current byte in theMessage to extract
    bool result = true;//false if theMessage was malformed

    //### Clear all fields ###
    for (unsigned int i = 0; i < protocolOrder.size(); i++)
        protocolOrder.at(i)->clear();

    while ((result) && (fieldIndex < protocolOrder.size()))
    {
        currentField = protocolOrder.at(fieldIndex);
        if (!currentField->parse(theMessage, byteIndex, errMsg))
            result = false;

        fieldIndex++;
    }//while result && fieldIndex

    if (byteIndex != theMessage.size())
    {
        errMsg = "There are extra bytes at the end of the message";
        result = false;
    }//if byteIndex

    return result;
}//parseMessage

DBuffer DProtocol::generateMessage(void)
{
    DField *currentField = NULL;//the current field we're adding to the message
    DBuffer result;//the message to return

    for (unsigned int i = 0; i < protocolOrder.size(); i++)
    {
        currentField = protocolOrder.at(i);
        currentField->generateMessage(result);
    }//for i

    return result;
}//generateMessage

void DProtocol::appendField(DField *newField)
{
    pair<map<string, DField*>::iterator, bool> ret;//the return value from the insert operation
    string fieldName = newField->name();

    ret = fields.insert(pair<string, DField*>(fieldName, newField));
    if (!ret.second)
    {
        DEngine::addFatal("DProtocol::addField(): field ");
        DEngine::addFatal(fieldName);
        DEngine::fatalError(" already exists");
    }//if !second

    protocolOrder.push_back(newField);
}//appendField

DField* DProtocol::getField(const string &fieldName)
{
    std::map<string, DField*>::iterator fieldIter;//points to the desired field in fields

    fieldIter = fields.find(fieldName);
    if (fieldIter == fields.end())
    {
        DEngine::addFatal("DProtocol::getField(): couldn't find ");
        DEngine::fatalError(fieldName);
    }//if fieldIter

    return fieldIter->second;
}//getField

bool DProtocol::getNextBytes(const DBuffer &theMessage,
                             unsigned int &byteIndex,
                             DBuffer &requestedSegment,
                             unsigned int numBytes)
{
    bool result = true;//false if there weren't enough bytes in theMessage

    requestedSegment.clear();

    if (byteIndex + numBytes > theMessage.size())
        result = false;
    else
    {
        requestedSegment.append(theMessage, byteIndex, numBytes);
        byteIndex += numBytes;
    }//else byteIndex + numBytes

    return result;
}//getNextBytes

bool DProtocol::extractLength(const DBuffer &theMessage,
                              unsigned int &byteIndex,
                              uint64_t &theLength,
                              string &errorDesc)
{
    DBuffer currentSegment;//the current length-related segment to extract
    uint32_t size32 = 0;//these contain an integer version of lengthBytes
    uint16_t size16 = 0;
    uint8_t size8 = 0;
    uint8_t rangeSize = 0;//the number of bytes in the number-of-values field
    bool result = true;//false if lengthBytes has an invalid size

    //### Extract the number of bytes in the length ###
    if (getNextBytes(theMessage, byteIndex, currentSegment, 1))
        rangeSize = currentSegment.getByte(0);
    else
    {
        errorDesc = "Expected a range size.";
        result = false;
    }//else getNextBytes()

    if (rangeSize == 0)
        theLength = 0;
    else
    {
        //### Extract the length ###
        if (result)
        {
            if (!getNextBytes(theMessage, byteIndex, currentSegment, rangeSize))
            {
                errorDesc = "Expected a field length.";
                result = false;
            }//if !getNextBytes
        }//if result

        //### Convert the length to a uint64_t ###
        if (result)
        {
            switch (currentSegment.size())
            {
                case 1:
                    currentSegment.extractVariable(0, size8);
                    theLength = size8;
                break;

                case 2:
                    currentSegment.extractVariable(0, size16);
                    theLength = size16;
                break;

                case 4:
                    currentSegment.extractVariable(0, size32);
                    theLength = size32;
                break;

                case 8:
                    currentSegment.extractVariable(0, theLength);
                break;

                default:
                    errorDesc = "Invalid size for the length value: ";
                    errorDesc.append(DEngine::intToStr(rangeSize));
                    result = false;
                break;
            }//switch size()
        }//if result
    }//else rangeSize

    return result;
}//extractLength

void DProtocol::appendSize(DBuffer &theMessage,
                           uint64_t theSize)
{
    uint32_t fourByteCount = 0;//the largest number we can hold in 4 unsigned bytes

    //We can't define unsigned constants larger than 2^31
    fourByteCount = 2147483647;
    fourByteCount *= 2;
    fourByteCount++;

    if (theSize == 0)
        theMessage.appendByte(0);

    else if (theSize <= 255)
    {
        theMessage.appendByte(1);
        theMessage.appendByte(static_cast<uint8_t>(theSize));
    }//if theData.size()

    else if (theSize <= 65535)
    {
        theMessage.appendByte(2);
        theMessage.appendVariable(static_cast<uint16_t>(theSize));
    }//else if theData.size()

    else if (theSize <= fourByteCount)
    {
        theMessage.appendByte(4);
        theMessage.appendVariable(static_cast<uint32_t>(theSize));
    }//else if size()

    else
    {
        theMessage.appendByte(8);
        theMessage.appendVariable(theSize);
    }//else size()
}//appendSize

string DProtocol::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
