/*
 * Copyright (C) 2012 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    DProtocol lets us build complex network messages. It's a vector of DFields,
    with each element being another part of the message. We can quickly compose messages
    consisting of primitives, std::strings and DBuffers. To save bandwidth, each
    element's type is never sent across the network. Instead, we rely on the client and server
    to define identical protocols and pre-agree on the message format.
*/

#ifndef DPROTOCOL_HPP_INCLUDED
#define DPROTOCOL_HPP_INCLUDED

#include "DEngine.hpp"

namespace libdasyne
{
    class DField;
    class DBuffer;

    class DProtocol
    {
        public:
            //constructor
            DProtocol(void);

            //destructor
            ~DProtocol(void);

            //Adds newField to the protocol. DProtocol takes ownership of it, and will
            //delete it when finished. newField should have a unique name, which we will use
            //to refer to it in getField()
            void appendField(DField *newField);

            //Return the specified field
            DField* getField(const std::string &fieldName);

            //Parse a message received from the network. theMessage should contain the complete message
            //we're expecting. Returns false if theMessage was malformed, and sets errMsg with a description
            //of the problem.
            bool parseMessage(const DBuffer &theMessage);

            //Generate a network message based on the current values in the DProtocol's fields.
            DBuffer generateMessage(void);

            //Returns a description of the most recent problem with this DProtocol
            std::string getErrMsg(void);

            //Adds a size to the message and a single byte before the size, containing the
            //size's length in bytes. This will be either using 0, 1, 2, 4 or 8 bytes.
            static void appendSize(DBuffer &theMessage,
                                   uint64_t theSize);

            //Extracts the next numBytes bytes from theMessage, and places them in requestedSegment.
            //requestedSegment is cleared first. We start copying from byteIndex, and increment byteIndex
            //by numBytes afterwards. Returns false if there weren't enough bytes in theMessage.
            static bool getNextBytes(const DBuffer &theMessage,
                                     unsigned int &byteIndex,
                                     DBuffer &requestedSegment,
                                     unsigned int numBytes);

            //Converts bytes in theMessage into a length value. It reads one byte to determine the number
            //the number of bytes in theLength, then reads those bytes and converts them to a uint64_t.
            //Returns false on failure and sets errorDesc with a description of the problem.
            static bool extractLength(const DBuffer &theMessage,
                                      unsigned int &byteIndex,
                                      uint64_t &theLength,
                                      std::string &errorDesc);

        private:
            //Each field from the protocol, ordered by name
            std::map<std::string, DField*> fields;

            //Each field from the protocol, ordered by their position in the message.
            //Contains pointers to elements in fields, don't delete.
            std::vector<DField*> protocolOrder;

            //A description of the most recent problem
            std::string errMsg;
    };//DProtocol
}//namespace libdasyne

#endif
